@@ -43,6 +43,7 @@
 #include <apr_atomic.h>
 
 #include "serf.h"
+#include "serf_private.h"
 #include "serf_bucket_util.h"
 
 #include <openssl/bio.h>
@@ -64,8 +65,6 @@
 #endif
 
 
-/*#define SSL_VERBOSE*/
-
 /*
  * Here's an overview of the SSL bucket's relationship to OpenSSL and serf.
  *
@@ -202,6 +201,47 @@ struct serf_ssl_certificate_t {
     int depth;
 };
 
+static void disable_compression(serf_ssl_context_t *ssl_ctx);
+
+#if SSL_VERBOSE
+/* Log all ssl alerts that we receive from the server. */
+static void
+apps_ssl_info_callback(const SSL *s, int where, int ret)
+{
+    const char *str;
+    int w;
+    w = where & ~SSL_ST_MASK;
+
+    if (w & SSL_ST_CONNECT)
+        str = "SSL_connect";
+    else if (w & SSL_ST_ACCEPT)
+        str = "SSL_accept";
+    else
+        str = "undefined";
+
+    if (where & SSL_CB_LOOP) {
+        serf__log(SSL_VERBOSE, __FILE__, "%s:%s\n", str,
+                  SSL_state_string_long(s));
+    }
+    else if (where & SSL_CB_ALERT) {
+        str = (where & SSL_CB_READ) ? "read" : "write";
+        serf__log(SSL_VERBOSE, __FILE__, "SSL3 alert %s:%s:%s\n",
+               str,
+               SSL_alert_type_string_long(ret),
+               SSL_alert_desc_string_long(ret));
+    }
+    else if (where & SSL_CB_EXIT) {
+        if (ret == 0)
+            serf__log(SSL_VERBOSE, __FILE__, "%s:failed in %s\n", str,
+                      SSL_state_string_long(s));
+        else if (ret < 0) {
+            serf__log(SSL_VERBOSE, __FILE__, "%s:error in %s\n", str,
+                      SSL_state_string_long(s));
+        }
+    }
+}
+#endif
+
 /* Returns the amount read. */
 static int bio_bucket_read(BIO *bio, char *in, int inlen)
 {
@@ -210,17 +250,15 @@ static int bio_bucket_read(BIO *bio, char *in, int inlen)
     apr_status_t status;
     apr_size_t len;
 
-#ifdef SSL_VERBOSE
-    printf("bio_bucket_read called for %d bytes\n", inlen);
-#endif
+    serf__log(SSL_VERBOSE, __FILE__, "bio_bucket_read called for %d bytes\n",
+              inlen);
 
     if (ctx->encrypt.status == SERF_ERROR_WAIT_CONN
         && BIO_should_read(ctx->bio)) {
-#ifdef SSL_VERBOSE
-        printf("bio_bucket_read waiting: (%d %d %d)\n",
+        serf__log(SSL_VERBOSE, __FILE__,
+                  "bio_bucket_read waiting: (%d %d %d)\n",
            BIO_should_retry(ctx->bio), BIO_should_read(ctx->bio),
            BIO_get_retry_flags(ctx->bio));
-#endif
         /* Falling back... */
         ctx->encrypt.exhausted_reset = 1;
         BIO_clear_retry_flags(bio);
@@ -229,9 +267,9 @@ static int bio_bucket_read(BIO *bio, char *in, int inlen)
     status = serf_bucket_read(ctx->decrypt.pending, inlen, &data, &len);
 
     ctx->decrypt.status = status;
-#ifdef SSL_VERBOSE
-    printf("bio_bucket_read received %d bytes (%d)\n", len, status);
-#endif
+
+    serf__log(SSL_VERBOSE, __FILE__, "bio_bucket_read received %d bytes (%d)\n",
+              len, status);
 
     if (!SERF_BUCKET_READ_ERROR(status)) {
         /* Oh suck. */
@@ -254,16 +292,15 @@ static int bio_bucket_write(BIO *bio, const char *in, int inl)
     serf_ssl_context_t *ctx = bio->ptr;
     serf_bucket_t *tmp;
 
-#ifdef SSL_VERBOSE
-    printf("bio_bucket_write called for %d bytes\n", inl);
-#endif
+    serf__log(SSL_VERBOSE, __FILE__, "bio_bucket_write called for %d bytes\n",
+              inl);
+
     if (ctx->encrypt.status == SERF_ERROR_WAIT_CONN
         && !BIO_should_read(ctx->bio)) {
-#ifdef SSL_VERBOSE
-        printf("bio_bucket_write waiting: (%d %d %d)\n",
+        serf__log(SSL_VERBOSE, __FILE__,
+                  "bio_bucket_write waiting: (%d %d %d)\n",
            BIO_should_retry(ctx->bio), BIO_should_read(ctx->bio),
            BIO_get_retry_flags(ctx->bio));
-#endif
         /* Falling back... */
         ctx->encrypt.exhausted_reset = 1;
         BIO_clear_retry_flags(bio);
@@ -428,6 +465,9 @@ validate_server_certificate(int cert_valid, X509_STORE_CTX *store_ctx)
             case X509_V_ERR_INVALID_CA:
                     failures |= SERF_SSL_CERT_UNKNOWNCA;
                     break;
+            case X509_V_ERR_CERT_REVOKED:
+                    failures |= SERF_SSL_CERT_REVOKED;
+                    break;
             default:
                     failures |= SERF_SSL_CERT_UNKNOWN_FAILURE;
                     break;
@@ -459,9 +499,13 @@ validate_server_certificate(int cert_valid, X509_STORE_CTX *store_ctx)
                                            failures, cert);
         if (status == APR_SUCCESS)
             cert_valid = 1;
-        else
+        else {
+            /* Even if openssl found the certificate valid, the application
+               told us to reject it. */
+            cert_valid = 0;
             /* Pass the error back to the caller through the context-run. */
             ctx->pending_err = status;
+        }
         apr_pool_destroy(subpool);
     }
 
@@ -494,7 +538,7 @@ validate_server_certificate(int cert_valid, X509_STORE_CTX *store_ctx)
             certs_len = 1;
         } else {
             int i;
-        
+
             certs_len = sk_X509_num(chain);
 
             /* Room for all the certs and a trailing NULL.  */
@@ -518,6 +562,9 @@ validate_server_certificate(int cert_valid, X509_STORE_CTX *store_ctx)
         if (status == APR_SUCCESS) {
             cert_valid = 1;
         } else {
+            /* Even if openssl found the certificate valid, the application
+               told us to reject it. */
+            cert_valid = 0;
             /* Pass the error back to the caller through the context-run. */
             ctx->pending_err = status;
         }
@@ -525,6 +572,15 @@ validate_server_certificate(int cert_valid, X509_STORE_CTX *store_ctx)
         apr_pool_destroy(subpool);
     }
 
+    /* Return a specific error if the server certificate is not accepted by
+       OpenSSL and the application has not set callbacks to override this. */
+    if (!cert_valid &&
+        !ctx->server_cert_chain_callback &&
+        !ctx->server_cert_callback)
+    {
+        ctx->pending_err = SERF_ERROR_SSL_CERT_FAILED;
+    }
+        
     return cert_valid;
 }
 
@@ -541,18 +597,15 @@ static apr_status_t ssl_decrypt(void *baton, apr_size_t bufsize,
     if (ctx->fatal_err)
         return ctx->fatal_err;
 
-#ifdef SSL_VERBOSE
-    printf("ssl_decrypt: begin %d\n", bufsize);
-#endif
+    serf__log(SSL_VERBOSE, __FILE__, "ssl_decrypt: begin %d\n", bufsize);
 
     /* Is there some data waiting to be read? */
     ssl_len = SSL_read(ctx->ssl, buf, bufsize);
     if (ssl_len > 0) {
-#ifdef SSL_VERBOSE
-        printf("ssl_decrypt: %d bytes (%d); status: %d; flags: %d\n",
-               ssl_len, bufsize, ctx->decrypt.status,
-               BIO_get_retry_flags(ctx->bio));
-#endif
+        serf__log(SSL_VERBOSE, __FILE__,
+                  "ssl_decrypt: %d bytes (%d); status: %d; flags: %d\n",
+                  ssl_len, bufsize, ctx->decrypt.status,
+                  BIO_get_retry_flags(ctx->bio));
         *len = ssl_len;
         return APR_SUCCESS;
     }
@@ -562,10 +615,9 @@ static apr_status_t ssl_decrypt(void *baton, apr_size_t bufsize,
     if (!SERF_BUCKET_READ_ERROR(status) && priv_len) {
         serf_bucket_t *tmp;
 
-#ifdef SSL_VERBOSE
-        printf("ssl_decrypt: read %d bytes (%d); status: %d\n", priv_len,
-               bufsize, status);
-#endif
+        serf__log(SSL_VERBOSE, __FILE__,
+                  "ssl_decrypt: read %d bytes (%d); status: %d\n",
+                  priv_len, bufsize, status);
 
         tmp = serf_bucket_simple_copy_create(data, priv_len,
                                              ctx->decrypt.pending->allocator);
@@ -580,9 +632,12 @@ static apr_status_t ssl_decrypt(void *baton, apr_size_t bufsize,
             switch (ssl_err) {
             case SSL_ERROR_SYSCALL:
                 *len = 0;
+                /* Return the underlying network error that caused OpenSSL
+                   to fail. ### This can be a crypt error! */
                 status = ctx->decrypt.status;
                 break;
             case SSL_ERROR_WANT_READ:
+            case SSL_ERROR_WANT_WRITE:
                 *len = 0;
                 status = APR_EAGAIN;
                 break;
@@ -616,10 +671,9 @@ static apr_status_t ssl_decrypt(void *baton, apr_size_t bufsize,
                 necessary mean the connection is closed, let's close
                 it here anyway.
                 We can optimize this later. */
-#ifdef SSL_VERBOSE
-                printf("ssl_decrypt: SSL read error: server shut down"\
-                       "connection!\n");
-#endif
+                serf__log(SSL_VERBOSE, __FILE__, 
+                          "ssl_decrypt: SSL read error: server"
+                          " shut down connection!\n");
                 status = APR_EOF;
             } else {
                 /* A fatal error occurred. */
@@ -627,18 +681,17 @@ static apr_status_t ssl_decrypt(void *baton, apr_size_t bufsize,
             }
         } else {
             *len = ssl_len;
-#ifdef SSL_VERBOSE
-            printf("---\n%s\n-(%d)-\n", buf, *len);
-#endif
+            serf__log(SSL_MSG_VERBOSE, __FILE__, 
+                      "---\n%.*s\n-(%d)-\n", *len, buf, *len);
         }
     }
     else {
         *len = 0;
     }
-#ifdef SSL_VERBOSE
-    printf("ssl_decrypt: %d %d %d\n", status, *len,
-           BIO_get_retry_flags(ctx->bio));
-#endif
+    serf__log(SSL_VERBOSE, __FILE__, 
+              "ssl_decrypt: %d %d %d\n", status, *len,
+              BIO_get_retry_flags(ctx->bio));
+
     return status;
 }
 
@@ -654,9 +707,7 @@ static apr_status_t ssl_encrypt(void *baton, apr_size_t bufsize,
     if (ctx->fatal_err)
         return ctx->fatal_err;
 
-#ifdef SSL_VERBOSE
-    printf("ssl_encrypt: begin %d\n", bufsize);
-#endif
+    serf__log(SSL_VERBOSE, __FILE__, "ssl_encrypt: begin %d\n", bufsize);
 
     /* Try to read already encrypted but unread data first. */
     status = serf_bucket_read(ctx->encrypt.pending, bufsize, &data, len);
@@ -670,29 +721,28 @@ static apr_status_t ssl_encrypt(void *baton, apr_size_t bufsize,
         if (APR_STATUS_IS_EOF(status)) {
             status = APR_SUCCESS;
         }
-#ifdef SSL_VERBOSE
-        printf("ssl_encrypt: %d %d %d (quick read)\n", status, *len,
-               BIO_get_retry_flags(ctx->bio));
-#endif
+
+        serf__log(SSL_VERBOSE, __FILE__, "ssl_encrypt: %d %d %d (quick read)\n",
+                  status, *len, BIO_get_retry_flags(ctx->bio));
+
         return status;
     }
 
     if (BIO_should_retry(ctx->bio) && BIO_should_write(ctx->bio)) {
-#ifdef SSL_VERBOSE
-        printf("ssl_encrypt: %d %d %d (should write exit)\n", status, *len,
-               BIO_get_retry_flags(ctx->bio));
-#endif
+        serf__log(SSL_VERBOSE, __FILE__,
+                  "ssl_encrypt: %d %d %d (should write exit)\n",
+                  status, *len, BIO_get_retry_flags(ctx->bio));
+
         return APR_EAGAIN;
     }
 
     /* If we were previously blocked, unblock ourselves now. */
     if (BIO_should_read(ctx->bio)) {
-#ifdef SSL_VERBOSE
-        printf("ssl_encrypt: reset %d %d (%d %d %d)\n", status,
-               ctx->encrypt.status,
-           BIO_should_retry(ctx->bio), BIO_should_read(ctx->bio),
-           BIO_get_retry_flags(ctx->bio));
-#endif
+        serf__log(SSL_VERBOSE, __FILE__, "ssl_encrypt: reset %d %d (%d %d %d)\n",
+                  status, ctx->encrypt.status,
+                  BIO_should_retry(ctx->bio), BIO_should_read(ctx->bio),
+                  BIO_get_retry_flags(ctx->bio));
+
         ctx->encrypt.status = APR_SUCCESS;
         ctx->encrypt.exhausted_reset = 0;
     }
@@ -734,34 +784,42 @@ static apr_status_t ssl_encrypt(void *baton, apr_size_t bufsize,
                 interim_bufsize -= vecs_data_len;
                 interim_len = vecs_data_len;
 
-#ifdef SSL_VERBOSE
-                printf("ssl_encrypt: bucket read %d bytes; status %d\n",
-                       interim_len, status);
-                printf("---\n%s\n-(%d)-\n", vecs_data, interim_len);
-#endif
+                serf__log(SSL_VERBOSE, __FILE__,
+                          "ssl_encrypt: bucket read %d bytes; "\
+                          "status %d\n", interim_len, status);
+                serf__log(SSL_MSG_VERBOSE, __FILE__, "---\n%.*s\n-(%d)-\n",
+                          interim_len, vecs_data, interim_len);
+
                 /* Stash our status away. */
                 ctx->encrypt.status = status;
 
                 ssl_len = SSL_write(ctx->ssl, vecs_data, interim_len);
-#ifdef SSL_VERBOSE
-                printf("ssl_encrypt: SSL write: %d\n", ssl_len);
-#endif
-                /* We're done. */
-                serf_bucket_mem_free(ctx->allocator, vecs_data);
+
+                serf__log(SSL_VERBOSE, __FILE__, 
+                          "ssl_encrypt: SSL write: %d\n", ssl_len);
 
                 /* If we failed to write... */
                 if (ssl_len < 0) {
                     int ssl_err;
 
-                    /* Ah, bugger. We need to put that data back. */
-                    serf_bucket_aggregate_prepend_iovec(ctx->encrypt.stream,
-                                                        vecs, vecs_read);
+                    /* Ah, bugger. We need to put that data back.
+                       Note: use the copy here, we do not own the original iovec
+                       data buffer so it will be freed on next read. */
+                    serf_bucket_t *vecs_copy =
+                        serf_bucket_simple_own_create(vecs_data,
+                                                      vecs_data_len,
+                                                      ctx->allocator);
+                    serf_bucket_aggregate_prepend(ctx->encrypt.stream,
+                                                  vecs_copy);
 
                     ssl_err = SSL_get_error(ctx->ssl, ssl_len);
-#ifdef SSL_VERBOSE
-                    printf("ssl_encrypt: SSL write error: %d\n", ssl_err);
-#endif
+
+                    serf__log(SSL_VERBOSE, __FILE__, 
+                              "ssl_encrypt: SSL write error: %d\n", ssl_err);
+
                     if (ssl_err == SSL_ERROR_SYSCALL) {
+                        /* Return the underlying network error that caused OpenSSL
+                           to fail. ### This can be a decrypt error! */
                         status = ctx->encrypt.status;
                         if (SERF_BUCKET_READ_ERROR(status)) {
                             return status;
@@ -773,12 +831,17 @@ static apr_status_t ssl_encrypt(void *baton, apr_size_t bufsize,
                             status = SERF_ERROR_WAIT_CONN;
                         }
                         else {
-                            ctx->fatal_err = status = SERF_ERROR_SSL_COMM_FAILED;
+                            ctx->fatal_err = status =
+                                SERF_ERROR_SSL_COMM_FAILED;
                         }
                     }
-#ifdef SSL_VERBOSE
-                    printf("ssl_encrypt: SSL write error: %d %d\n", status, *len);
-#endif
+
+                    serf__log(SSL_VERBOSE, __FILE__, 
+                              "ssl_encrypt: SSL write error: %d %d\n",
+                              status, *len);
+                } else {
+                    /* We're done with this data. */
+                    serf_bucket_mem_free(ctx->allocator, vecs_data);
                 }
             }
         }
@@ -805,10 +868,9 @@ static apr_status_t ssl_encrypt(void *baton, apr_size_t bufsize,
             *len += vecs[i].iov_len;
         }
 
-#ifdef SSL_VERBOSE
-        printf("ssl_encrypt read agg: %d %d %d %d\n", status, agg_status,
-               ctx->encrypt.status, *len);
-#endif
+        serf__log(SSL_VERBOSE, __FILE__,
+                  "ssl_encrypt read agg: %d %d %d %d\n", status, agg_status,
+            ctx->encrypt.status, *len);
 
         if (!agg_status) {
             status = agg_status;
@@ -821,11 +883,11 @@ static apr_status_t ssl_encrypt(void *baton, apr_size_t bufsize,
         ctx->encrypt.status = SERF_ERROR_WAIT_CONN;
     }
 
-#ifdef SSL_VERBOSE
-    printf("ssl_encrypt finished: %d %d (%d %d %d)\n", status, *len,
-           BIO_should_retry(ctx->bio), BIO_should_read(ctx->bio),
-           BIO_get_retry_flags(ctx->bio));
-#endif
+    serf__log(SSL_VERBOSE, __FILE__,
+              "ssl_encrypt finished: %d %d (%d %d %d)\n", status, *len,
+              BIO_should_retry(ctx->bio), BIO_should_read(ctx->bio),
+              BIO_get_retry_flags(ctx->bio));
+
     return status;
 }
 
@@ -911,6 +973,20 @@ static void init_ssl_libraries(void)
 #if APR_HAS_THREADS
         int i, numlocks;
 #endif
+
+#ifdef SSL_VERBOSE
+        /* Warn when compile-time and run-time version of OpenSSL differ in
+           major/minor version number. */
+        long libver = SSLeay();
+
+        if ((libver ^ OPENSSL_VERSION_NUMBER) & 0xFFF00000) {
+            serf__log(SSL_VERBOSE, __FILE__,
+                      "Warning: OpenSSL library version mismatch, compile-time "
+                      "was %lx, runtime is %lx.\n",
+                      OPENSSL_VERSION_NUMBER, libver);
+        }
+#endif
+
         CRYPTO_malloc_init();
         ERR_load_crypto_strings();
         SSL_load_error_strings();
@@ -971,7 +1047,7 @@ static int ssl_need_client_cert(SSL *ssl, X509 **cert, EVP_PKEY **pkey)
         }
 
         if (status || !cert_path) {
-          break;
+            break;
         }
 
         /* Load the x.509 cert file stored in PKCS12 */
@@ -1155,6 +1231,8 @@ static serf_ssl_context_t *ssl_init_context(void)
     SSL_CTX_set_verify(ssl_ctx->ctx, SSL_VERIFY_PEER,
                        validate_server_certificate);
     SSL_CTX_set_options(ssl_ctx->ctx, SSL_OP_ALL);
+    /* Disable SSL compression by default. */
+    disable_compression(ssl_ctx);
 
     ssl_ctx->ssl = SSL_new(ssl_ctx->ctx);
     ssl_ctx->bio = BIO_new(&bio_bucket_method);
@@ -1166,6 +1244,10 @@ static serf_ssl_context_t *ssl_init_context(void)
 
     SSL_set_app_data(ssl_ctx->ssl, ssl_ctx);
 
+#if SSL_VERBOSE
+    SSL_CTX_set_info_callback(ssl_ctx->ctx, apps_ssl_info_callback);
+#endif
+
     ssl_ctx->encrypt.stream = NULL;
     ssl_ctx->encrypt.stream_next = NULL;
     ssl_ctx->encrypt.pending = serf_bucket_aggregate_create(allocator);
@@ -1254,7 +1336,7 @@ apr_status_t serf_ssl_load_cert_file(
     apr_pool_t *pool)
 {
     FILE *fp = fopen(file_path, "r");
-        
+
     if (fp) {
         X509 *ssl_cert = PEM_read_X509(fp, NULL, NULL, NULL);
         fclose(fp);
@@ -1560,6 +1642,31 @@ const char *serf_ssl_cert_export(
     return encoded_cert;
 }
 
+/* Disables compression for all SSL sessions. */
+static void disable_compression(serf_ssl_context_t *ssl_ctx)
+{
+#ifdef SSL_OP_NO_COMPRESSION
+    SSL_CTX_set_options(ssl_ctx->ctx, SSL_OP_NO_COMPRESSION);
+#endif
+}
+
+apr_status_t serf_ssl_use_compression(serf_ssl_context_t *ssl_ctx, int enabled)
+{
+    if (enabled) {
+#ifdef SSL_OP_NO_COMPRESSION
+        SSL_clear_options(ssl_ctx->ssl, SSL_OP_NO_COMPRESSION);
+        return APR_SUCCESS;
+#endif
+    } else {
+#ifdef SSL_OP_NO_COMPRESSION
+        SSL_set_options(ssl_ctx->ssl, SSL_OP_NO_COMPRESSION);
+        return APR_SUCCESS;
+#endif
+    }
+
+    return APR_EGENERAL;
+}
+
 static void serf_ssl_destroy_and_data(serf_bucket_t *bucket)
 {
     ssl_context_t *ctx = bucket->data;
